```
# ~~~
# This file is part of PhD-thesis:
#
#           "Adaptive Reduced Basis Methods for Multiscale Problems
#               and Large-scale PDE-constrained Optimization"
#
# by: Tim Keil
#
#   https://github.com/TiKeil/Supplementary-Material-for-PhD-thesis
#
# Copyright 2019-2022 all developers. All rights reserved.
# License: Licensed as BSD 2-Clause License (http://opensource.org/licenses/BSD-2-Clause)
# Authors:
#   Tim Keil (2022)
# ~~~
```

In this project, we provide information and code for the numerical experiments in the PhD thesis "Adaptive Reduced Basis Methods for Multiscale Problems and Large-scale PDE-constrained Optimization" by Tim Keil.
We refer to the repository in https://github.com/TiKeil/Supplementary-Material-for-PhD-thesis.

# How to quickly find the code used for the respective Experiments in the thesis

Most of the code used for the thesis has been published along with the papers, and corresponding GitHub repositories are available.
Apart from these, this repository aims to provide missing code and results that were used in the thesis, see below.
In the following, we itemize each of the numbered experiments:

- Experiment 1: The proof-of-concept experiment is part of this repository, see below.
- Experiment 2, 3 and 4: see [this repository](https://github.com/TiKeil/NCD-corrected-TR-RB-approach-for-pde-opt)
- Experiment 5 and 6: see [this repository](https://github.com/TiKeil/Proj-Newton-NCD-corrected-TR-RB-for-pde-opt)
- Experiment 7 and 8: see [this repository](https://github.com/TiKeil/Petrov-Galerkin-TR-RB-for-pde-opt)
- Experiment 9: see [this repository](https://github.com/gridlod-community/gridlod-on-perturbations-super)
- Experiment 10 and 11: see [this repository](https://github.com/TiKeil/Two-scale-RBLOD)
- Experiment 12 and 13: see [this repository](https://github.com/TiKeil/Trust-region-TSRBLOD-code)

For more insights, we refer to the respective repositories.

# Unpublished code that is part of this repository

Two additional experiments are provided:

1) The proof-of-concept experiment that is used as motivation in Section 2.5.6, as illustration of the TR-RB algorithm in Section 3.5.3, and as the test problem for further approaches in Section 3.6.6.

2) The TR-RBLOD results for Experiment 12 and 13 as well as the problem specifics for Experiment 14, the comparison between the TR-TSRBLOD and the TR-RBLOD.

For just taking a look at the experiment outputs and data, you do not need to
install the software. Just go to `notebooks/` for taking a look at 1) and to `scripts/test_outputs/` for an extended overview concerning 2)

If you want to have a closer look at the implementation or generate the results by
yourself, we provide simple setup instructions for configuring your own Python environment.
We note that our setup instructions are written for Ubuntu Linux only and we do not provide
setup instructions for MacOS and Windows.
Our setup instructions have successfully been tested on a fresh Ubuntu 20.04.2.0 LTS system.
The actual experiments for 2) have been computed on the
[PALMA II HPC cluster](<https://www.uni-muenster.de/IT/en/services/unterstuetzungsleistung/hpc/index.shtml>).

# Organization of the repository

We used several external software packages:

- [pyMOR](https://pymor.org) is a software library for Model Order Reduction.
- [gridlod](https://github.com/fredrikhellman/gridlod) is a discretization toolkit for the
Localized Orthogonal Decompostion (LOD) method. 
- [perturbations-for-2d-data](https://github.com/TiKeil/perturbations-for-2d-data) contains
a coefficient generator for constructing randomized and highly oscillating coefficients.
- [TSRBLOD](https://github.com/TiKeil/Two-scale-RBLOD) this code has been used for the [paper](https://arxiv.org/abs/2111.08643) where the TSRBLOD has been introduced. The essential code for our project can be found the respective `rblod` module.

For the TR-TSRBLOD, we have developed a Python module `pdeopt` that provides all other code that was required for the project. We also note that parts of this code have already been developed in the respective TR-RB software.
The rest of the code is contained in `scripts`, where you find the main scripts for the numerical experiments.

# Setup

On a standard Ubuntu system (with Python and C compilers installed) it will most likely be enough
to just run our setup script. For that, please execute the provided setup script via 

```
cd Supplementary-Material-for-PhD-thesis
./setup.sh
```

If this does not work for you, and you don't know how to help yourself,
please follow the extended setup instructions below.

## Installation on a fresh system

We also provide setup instructions for a fresh Ubuntu system (20.04.2.0 LTS).
The following steps need to be taken:

```
sudo apt update
sudo apt upgrade
sudo apt install git
sudo apt install build-essential
sudo apt install python3-dev
sudo apt install python3-venv
sudo apt install libopenmpi-dev
sudo apt install libsuitesparse-dev
```

Now you are ready to clone the repository and run the setup script:

```
cd Supplementary-Material-for-PhD-thesis
./setup.sh
```

# Running the code

For the proof-of-concept experiment 1), just activate the venv and start a jupyter-notebook server
```
source venv/bin/activate
jupyter-notebook --notebook-dir=notebooks
```

We recommend to use notebook extensions for a better overview in the notebooks.
After starting the jupyter-notebook server go to Nbextensions, deactivate the first box and activate at least `codefolding` and `Collapsible Headings`.

For 2), you can make sure your that setup is complete by running the minimal test script
```
cd scripts/test_scripts
mpirun python minimal_test.py
```

If this works fine (with a summary of methods in the end), your setup is working well.

Moreover, in the respective `test_outputs` directory, all figures from the paper can directly be constructed with the respective files.
For executing Python scripts, you need to activate the virtual environment by

```
source venv/bin/activate
```

Note that the jobs require a HPC computing system.
In particular, starting the shell scripts with only a few parallel cores (or even without `mpirun`)
on your local computer may take days to weeks.

Please have a look at the description of the main scripts to try different configurations of the given problem classes.
Note that it is also possible to solve your own parameterized problems with our code since the problem definitions that are used in
`pdeopt/problems.py` are very general. 

# Questions

If there are any questions of any kind, please contact us via <tim.keil@wwu.de>.
